import yaml

from seml.utils.errors import ConfigError


class YamlUniqueLoader(yaml.FullLoader):
    """
    Custom YAML loader that disallows duplicate keys

    From https://github.com/encukou/naucse_render/commit/658197ed142fec2fe31574f1ff24d1ff6d268797
    Workaround for PyYAML issue: https://github.com/yaml/pyyaml/issues/165
    This disables some uses of YAML merge (`<<`)
    """


def construct_mapping(loader, node, deep=False):
    """Construct a YAML mapping node, avoiding duplicates"""
    loader.flatten_mapping(node)
    result = {}
    for key_node, value_node in node.value:
        key = loader.construct_object(key_node, deep=deep)
        if key in result:
            raise ConfigError(f"Found duplicate keys: '{key}'")
        result[key] = loader.construct_object(value_node, deep=deep)
    return result


YamlUniqueLoader.add_constructor(
    yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
    construct_mapping,
)



import logging
import shlex
import urllib
from typing import Optional

from seml.experiment.config import generate_named_configs
from seml.experiment.config import (
    resolve_interpolations as resolve_config_interpolations,
)
from seml.settings import SETTINGS
from seml.utils.errors import ArgumentError, MongoDBError
from seml.utils.network import find_free_port


def _generate_debug_attach_url(ip_address, port):
    import json

    launch_config = {
        'type': 'debugpy',
        'request': 'attach',
        'connect': {'host': ip_address, 'port': port},
        'pathMappings': [{'localRoot': '${workspaceFolder}', 'remoteRoot': '.'}],
    }
    launch_config = urllib.parse.quote(json.dumps(launch_config))
    return f'vscode://fabiospampinato.vscode-debug-launcher/launch?args={launch_config}'


def get_environment_variables(gpus=None, cpus=None, environment_variables=None):
    if environment_variables is None:
        environment_variables = {}

    if gpus is not None:
        if isinstance(gpus, list):
            raise ArgumentError(
                'Received an input of type list to set CUDA_VISIBLE_DEVICES. '
                'Please pass a string for input "gpus", '
                'e.g. "1,2" if you want to use GPUs with IDs 1 and 2.'
            )
        environment_variables['CUDA_VISIBLE_DEVICES'] = str(gpus)
    if cpus is not None:
        environment_variables['OMP_NUM_THREADS'] = str(cpus)
    return environment_variables


def get_config_overrides(config):
    return ' '.join(map(shlex.quote, config))


def get_shell_command(interpreter, exe, config, env: Optional[dict] = None):
    config_overrides = get_config_overrides(config)

    if env is None or len(env) == 0:
        return f'{interpreter} {exe} with {config_overrides}'
    else:
        env_overrides = ' '.join(
            f'{key}={shlex.quote(val)}' for key, val in env.items()
        )

        return f'{env_overrides} {interpreter} {exe} with {config_overrides}'


def value_to_string(value, use_json=False):
    from seml.utils.json import PythonEncoder

    # We need the json encoding for vscode due to https://github.com/microsoft/vscode/issues/91578
    # Once this bug has been fixed we should only rely on `repr` and remove this code.
    if use_json:
        return PythonEncoder().encode(value)
    else:
        return repr(value)


def get_command_from_exp(
    exp,
    db_collection_name,
    verbose=False,
    unobserved=False,
    post_mortem=False,
    debug=False,
    debug_server=False,
    print_info=True,
    use_json=False,
    unresolved=False,
    resolve_interpolations: bool = True,
):
    from seml.console import console

    if 'executable' not in exp['seml']:
        raise MongoDBError(
            f"No executable found for experiment {exp['_id']}. Aborting."
        )
    exe = exp['seml']['executable']

    if unresolved:
        config_unresolved = exp.get('config_unresolved', exp['config'])
        config, named_configs = tuple(
            zip(*generate_named_configs([config_unresolved]))
        )[0]
        # Variable interpolation in unresolved and named configs

        if resolve_interpolations:
            import uuid

            key_named_configs = str(uuid.uuid4())
            interpolated = resolve_config_interpolations(
                {
                    **exp,
                    'config_unresolved': config_unresolved,
                    key_named_configs: named_configs,
                },
                allow_interpolation_keys=list(SETTINGS.ALLOW_INTERPOLATION_IN)
                + ['config_unresolved', key_named_configs],
            )

            config = {
                k: v
                for k, v in interpolated['config_unresolved'].items()
                if not k.startswith(SETTINGS.NAMED_CONFIG.PREFIX)
            }
            named_configs = interpolated[key_named_configs]
        else:
            config = {
                k: v
                for k, v in config_unresolved.items()
                if not k.startswith(SETTINGS.NAMED_CONFIG.PREFIX)
            }
    else:
        assert (
            resolve_interpolations
        ), 'In resolved configs, interpolations are automatically resolved'
        config = exp['config']
        named_configs = []

    config['db_collection'] = db_collection_name
    if not unobserved:
        config['overwrite'] = exp['_id']

    # We encode values with `repr` such that we can decode them with `eval`. While `shlex.quote`
    # may cause messy commands with lots of single quotes JSON doesn't match Python 1:1, e.g.,
    # boolean values are lower case in JSON (true, false) but start with capital letters in Python.
    config_strings = [
        f'{key}={value_to_string(val, use_json)}' for key, val in config.items()
    ]
    config_strings += named_configs

    # TODO (?): Variable interpolation for unresolved CLI calls

    if not verbose:
        config_strings.append('--force')
    if unobserved:
        config_strings.append('--unobserved')
    if post_mortem:
        config_strings.append('--pdb')
    if debug:
        config_strings.append('--debug')

    if debug_server:
        ip_address, port = find_free_port()
        if print_info:
            logging.info(
                f"Starting debug server with IP '{ip_address}' and port '{port}'. "
                f'Experiment will wait for a debug client to attach.'
            )
            attach_link = _generate_debug_attach_url(ip_address, port)

            logging.info(
                "If you are using VSCode, you can use the 'Debug Launcher' extension to attach:"
            )
            console.out(attach_link)

        interpreter = (
            f'python -m debugpy --listen {ip_address}:{port} --wait-for-client'
        )
    else:
        interpreter = 'python'

    return interpreter, exe, config_strings



import logging
from copy import deepcopy

from seml.database import get_collection
from seml.settings import SETTINGS

States = SETTINGS.STATES

__all__ = ['get_results']


def parse_jsonpickle(db_entry):
    import json

    import jsonpickle
    import jsonpickle.ext.numpy as jsonpickle_numpy

    jsonpickle_numpy.register_handlers()
    try:
        p = jsonpickle.pickler.Pickler(keys=False)
        parsed = jsonpickle.loads(json.dumps(db_entry, default=p.flatten), keys=False)
    except IndexError:
        parsed = db_entry
    return parsed


def get_results(
    db_collection_name,
    fields=None,
    to_data_frame=False,
    mongodb_config=None,
    states=None,
    filter_dict=None,
    parallel: bool = False,
    progress: bool = True,
):
    """
    Get experiment results from the MongoDB.
    Parameters
    ----------
    db_collection_name: str
        Name of the MongoDB collection.
    fields: list (optional).
        Database attributes to extract. Default: ['config', 'result'].
    to_data_frame: bool, default: False
        Whether to convert the results into a Pandas DataFrame.
    mongodb_config: dict (optional)
        MongoDB credential dictionary. If None, uses the credentials specified by `seml configure`.
    states: list of strings (optional)
        Extract only experiments with certain states. Default: ['COMPLETED'].
    filter_dict: dict (optional)
        Custom dictionary for filtering results from the MongoDB.
    parallel: bool, default: False
        If True, unserialize entries in parallel. Use for very large experiment collections.
    progress: bool, default: False
        If True, show a progress bar.

    Returns
    -------

    """
    import functools

    import pandas as pd

    from seml.console import track

    if fields is None:
        fields = ['config', 'result']

    if states is None:
        states = States.COMPLETED

    if filter_dict is None:
        filter_dict = {}

    track = functools.partial(track, disable=not progress)

    collection = get_collection(
        db_collection_name,
        mongodb_config=mongodb_config,
    )

    if len(states) > 0:
        if 'status' in filter_dict:
            logging.warning(
                "'states' argument is not empty and will overwrite 'filter_dict['status']'."
            )
        filter_dict = deepcopy(filter_dict)
        filter_dict['status'] = {'$in': states}

    cursor = collection.find(filter_dict, fields)
    results = [x for x in track(cursor, total=collection.count_documents(filter_dict))]

    if parallel:
        from multiprocessing import Pool

        with Pool() as p:
            parsed = list(track(p.imap(parse_jsonpickle, results), total=len(results)))
    else:
        parsed = [parse_jsonpickle(entry) for entry in track(results)]
    if to_data_frame:
        parsed = pd.json_normalize(parsed, sep='.')
    return parsed



import logging
from typing import Dict, List, Optional

from seml.commands.manage import detect_killed
from seml.database import build_filter_dict, get_collection
from seml.experiment.description import resolve_description
from seml.settings import SETTINGS
from seml.utils import slice_to_str, to_slices
from seml.utils.errors import MongoDBError

States = SETTINGS.STATES


def collection_set_description(
    db_collection_name: str,
    description: str,
    sacred_id: Optional[int] = None,
    filter_states: Optional[List[str]] = None,
    filter_dict: Optional[Dict] = None,
    batch_id: Optional[int] = None,
    yes: bool = False,
    resolve: bool = True,
):
    """Sets (or updates) the description of experiment(s).

    Parameters
    ----------
    db_collection_name : str
        Name of the collection to delete descriptions from
    description : str
        The description to set.
    sacred_id : Optional[int], optional
        If given, the id of the experiment to delete the description of. Overrides other filters, by default None
    filter_states : Optional[List[str]], optional
        Filter on experiment states, by default None
    filter_dict : Optional[Dict], optional
        Additional filters, by default None
    batch_id : Optional[int], optional
        Filter on the batch ID of experiments, by default None
    yes : bool, optional
        Whether to override confirmation prompts, by default False
    resolve : bool, optional
        Whether to use omegaconf to resolve descriptions
    """
    from pymongo import UpdateOne

    from seml.console import prompt

    collection = get_collection(db_collection_name)

    filter_dict = build_filter_dict(
        filter_states, batch_id, filter_dict, sacred_id=sacred_id
    )
    exps = list(collection.find(filter_dict, {}))
    if len(exps) == 0 and sacred_id is not None:
        raise MongoDBError(f'No experiment found with ID {sacred_id}.')
    descriptions_resolved = {
        exp['_id']: resolve_description(description, exp) if resolve else description
        for exp in exps
    }
    num_to_overwrite = len(
        list(
            filter(
                lambda exp: exp.get('seml', {}).get(
                    'description', descriptions_resolved[exp['_id']]
                )
                != descriptions_resolved[exp['_id']],
                exps,
            )
        )
    )

    if (
        not yes
        and num_to_overwrite >= SETTINGS.CONFIRM_THRESHOLD.DESCRIPTION_UPDATE
        and not prompt(
            f'{num_to_overwrite} experiment(s) have a different description. Proceed?',
            type=bool,
        )
    ):
        exit(1)
    if len(list(filter(lambda exp: exp['status'] in States.RUNNING, exps))):
        logging.warn(
            f'Updating the description of {States.RUNNING[0]} experiments: This may not have an'
            ' effect, as sacred overwrites experiments with each tick.'
        )
    result = collection.bulk_write(
        [
            UpdateOne({'_id': _id}, {'$set': {'seml.description': description}})
            for _id, description in descriptions_resolved.items()
        ]
    )
    logging.info(f'Updated the descriptions of {result.modified_count} experiments.')


def collection_delete_description(
    db_collection_name: str,
    sacred_id: Optional[int] = None,
    filter_states: Optional[List[str]] = None,
    filter_dict: Optional[Dict] = None,
    batch_id: Optional[int] = None,
    yes: bool = False,
):
    """Deletes the description of experiments

    Parameters
    ----------
    db_collection_name : str
        Name of the collection to delete descriptions from
    sacred_id : Optional[int], optional
        If given, the id of the experiment to delete the descriptino of. Overrides other filters, by default None
    filter_states : Optional[List[str]], optional
        Filter on experiment states, by default None
    filter_dict : Optional[Dict], optional
        Additional filters, by default None
    batch_id : Optional[int], optional
        Filter on the batch ID of experiments, by default None
    yes : bool, optional
        Whether to override confirmation prompts, by default False
    """
    from seml.console import prompt

    collection = get_collection(db_collection_name)
    update = {'$unset': {'seml.description': ''}}
    filter_dict = build_filter_dict(
        filter_states, batch_id, filter_dict, sacred_id=sacred_id
    )
    exps = [
        exp
        for exp in collection.find(filter_dict, {'seml.description': 1})
        if exp.get('seml', {}).get('description', None) is not None
    ]
    if (
        not yes
        and len(exps) >= SETTINGS.CONFIRM_THRESHOLD.DESCRIPTION_DELETE
        and not prompt(
            f'Deleting descriptions of {len(exps)} experiment(s). Proceed?', type=bool
        )
    ):
        exit(1)
    result = collection.update_many(filter_dict, update)
    logging.info(f'Deleted the descriptions of {result.modified_count} experiments.')


def collection_list_descriptions(db_collection_name: str, update_status: bool = False):
    """Lists the descriptions of experiments

    Parameters
    ----------
    db_collection_name : str
        Name of the collection to list descriptions from
    update_status : bool
        Whether to detect killed experiments
    """
    from rich.align import Align

    from seml.console import Table, console

    collection = get_collection(db_collection_name)

    # Handle status updates
    if update_status:
        detect_killed(db_collection_name, print_detected=False)
    else:
        logging.warning(
            f'Status of {States.RUNNING[0]} experiments may not reflect if they have died or been canceled. Use the `--update-status` flag instead.'
        )

    description_slices = {
        (obj['_id'] if obj['_id'] else ''): {
            'ids': to_slices(obj['ids']),
            'batch_ids': to_slices(obj['batch_ids']),
            'states': set(obj['states']),
        }
        for obj in collection.aggregate(
            [
                {
                    '$group': {
                        '_id': '$seml.description',
                        'ids': {'$addToSet': '$_id'},
                        'batch_ids': {'$addToSet': '$batch_id'},
                        'states': {'$addToSet': '$status'},
                    }
                }
            ]
        )
    }

    table = Table(show_header=True)
    table.add_column('Description', justify='left')
    table.add_column('Experiment IDs', justify='left')
    table.add_column('Batch IDs', justify='left')
    table.add_column('Status', justify='left')
    for description in sorted(description_slices):
        slices = description_slices[description]
        table.add_row(
            description,
            ', '.join(map(slice_to_str, slices['ids'])),
            ', '.join(map(slice_to_str, slices['batch_ids'])),
            ', '.join(slices['states']),
        )
    console.print(Align(table, align='center'))



import logging

from seml.settings import SETTINGS


def prompt_ssh_forward():
    """
    Prompt the user for SSH Forward settings. The output format corresponds
    to the argument of sshtunnel.SSHTunnelForwarder.
    """
    from seml.console import prompt

    logging.info('Configuring SSH Forward settings.')
    ssh_host = prompt('SSH host')
    port = prompt('Port', default=22, type=int)
    username = prompt('User name')
    ssh_pkey = prompt('Path to SSH private key', default='~/.ssh/id_rsa')
    return dict(
        ssh_address_or_host=ssh_host,
        ssh_port=port,
        ssh_username=username,
        ssh_pkey=ssh_pkey,
    )


def mongodb_configure(setup_ssh_forward: bool = False):
    import yaml

    from seml.console import prompt

    if SETTINGS.DATABASE.MONGODB_CONFIG_PATH.exists() and not prompt(
        f'MongoDB configuration {SETTINGS.DATABASE.MONGODB_CONFIG_PATH} already exists and will be overwritten.\nContinue?',
        type=bool,
    ):
        return
    logging.info('Configuring MongoDB. Warning: Password will be stored in plain text.')
    host = prompt('MongoDB host')
    port = prompt('Port', default=27017, type=int)
    database = prompt('Database name')
    username = prompt('User name')
    password = prompt('Password', hide_input=True)
    file_path = SETTINGS.DATABASE.MONGODB_CONFIG_PATH
    config = dict(
        host=host,
        port=port,
        database=database,
        username=username,
        password=password,
    )
    if setup_ssh_forward:
        config['ssh_config'] = prompt_ssh_forward()
    config_string = yaml.dump(config)
    logging.info(
        f"Saving the following configuration to {file_path}:\n"
        f"{config_string.replace(f'{password}', '********')}"
    )
    file_path.parent.mkdir(parents=True, exist_ok=True)
    with open(file_path, 'w') as f:
        f.write(config_string)


def configure(
    all: bool = False, mongodb: bool = False, setup_ssh_forward: bool = False
):
    configured_any = False
    if mongodb or all:
        mongodb_configure(setup_ssh_forward=setup_ssh_forward)
        configured_any = True
    if not configured_any:
        logging.info('Did not specify any configuration to configure')



from pathlib import Path
from runpy import run_path

import seml.utils.typer as typer
from seml.utils import merge_dicts
from seml.utils.module_hider import ModuleHider

# The YAML, json import is rather slow
with ModuleHider(
    'yaml',
    'json',
    'simplejson',
    'importlib_metadata',
    'importlib.metadata',
):
    from munch import munchify

__all__ = ('SETTINGS',)

APP_DIR = Path(typer.get_app_dir('seml'))

SETTINGS = munchify(
    {
        # Location of user-specific settings.py file containing a SETTINGS dict.
        # With this dict you can change anything that is set here, conveniently from your home directory.
        # Default: $HOME/.config/seml/settings.py
        'USER_SETTINGS_PATH': APP_DIR / 'settings.py',
        # Directory which is used on the compute nodes to dump scripts and Python code.
        # Only change this if you know what you're doing.
        'TMP_DIRECTORY': '/tmp',
        'TEMPLATE_REMOTE': 'https://github.com/TUM-DAML/seml-templates.git',
        'CODE_CHECKPOINT_REMOVE_SRC_DIRECTORY': True,
        'DATABASE': {
            # location of the MongoDB config. Default: $HOME/.config/seml/monogdb.config
            'MONGODB_CONFIG_PATH': APP_DIR / 'mongodb.config'
        },
        'SLURM_DEFAULT': {
            'experiments_per_job': 1,
            'sbatch_options': {
                'time': '0-08:00',
                'nodes': 1,
                'cpus-per-task': 1,
                'mem': '8G',
            },
        },
        'SBATCH_OPTIONS_TEMPLATES': {
            # This is a special template used for `seml jupyter`
            'JUPYTER': {
                'cpus-per-task': 2,
                'mem': '16G',
                'gres': 'gpu:1',
                'qos': 'interactive',
                'job-name': 'jupyter',
                'output': 'jupyter-%j.out',
                'partition': 'gpu_gtx1080',
            },
            # Extend this with your custom templates.
            'GPU': {
                'cpus-per-task': 2,
                'mem': '16G',
                'gres': 'gpu:1',
            },
        },
        'STATES': {
            'STAGED': ['STAGED', 'QUEUED'],  # QUEUED for backward compatibility
            'PENDING': ['PENDING'],
            'RUNNING': ['RUNNING'],
            'FAILED': ['FAILED'],
            'KILLED': ['KILLED'],
            'INTERRUPTED': ['INTERRUPTED'],
            'COMPLETED': ['COMPLETED'],
        },
        'SLURM_STATES': {
            'PENDING': [
                'PENDING',
                'CONFIGURING',
                'REQUEUE_FED',
                'REQUEUE_HOLD',
                'REQUEUED',
                'RESIZING',
            ],
            'RUNNING': [
                'RUNNING',
                'SIGNALING',
            ],  # Python code can still be executed while in SIGNALING
            'PAUSED': ['STOPPED', 'SUSPENDED', 'SPECIAL_EXIT'],
            'INTERRUPTED': ['CANCELLED'],  # Caused by user command
            'FAILED': [
                'FAILED',
                'BOOT_FAIL',
                'DEADLINE',
                'NODE_FAIL',
                'OUT_OF_MEMORY',
                'PREEMPTED',
                'REVOKED',
                'TIMEOUT',
            ],
            # REVOKED is not failed, but would need code that handles multi-cluster operation
            'COMPLETED': ['COMPLETED', 'COMPLETING', 'STAGE_OUT'],
        },
        'VALID_SEML_CONFIG_VALUES': [
            'executable',
            'name',
            'output_dir',
            'conda_environment',
            'project_root_dir',
            'description',
            'stash_all_py_files',
        ],
        'SEML_CONFIG_VALUE_VERSION': 'version',
        'VALID_SLURM_CONFIG_VALUES': [
            'experiments_per_job',
            'max_simultaneous_jobs',
            'sbatch_options_template',
            'sbatch_options',
            'overrides',
        ],
        'LOGIN_NODE_NAMES': ['fs'],
        'OBSERVERS': {
            'NEPTUNE': {
                'AUTH_TOKEN': 'YOUR_AUTH_TOKEN',
            },
            'SLACK': {
                'WEBHOOK': 'YOUR_WEBHOOK',
            },
            'MATTERMOST': {
                'WEBHOOK': 'YOUR_WEBHOOK',
                'DEFAULT_CHANNEL': 'YOUR_DEFAULT_CHANNEL',
            },
        },
        'CONFIG_EXCLUDE_KEYS': [
            '__doc__'
        ],  # keys that will be excluded from resolved configurations, sacred for some reason captures the docstring attribute
        # Which key is treated as the experiment seed
        'CONFIG_KEY_SEED': 'seed',
        'ALLOW_INTERPOLATION_IN': [
            'seml.description',
            'config',
        ],  # in which fields to allow variable interpolation
        'NAMED_CONFIG': {
            'PREFIX': '+',  # prefix for all named configuration parameters
            'KEY_NAME': 'name',  # key that identifies the name of a named config
            'KEY_PRIORITY': 'priority',  # key that identifies the priority of a named config
        },
        'CONFIRM_THRESHOLD': {
            'DELETE': 10,
            'RESET': 10,
            'CANCEL': 10,
            'DESCRIPTION_DELETE': 10,
            'DESCRIPTION_UPDATE': 10,
        },
        'EXPERIMENT': {
            'CAPTURE_OUTPUT': False,  # whether to capture the output of the experiment in the database
            'TERMINAL_WIDTH': 80,  # width of the terminal for rich output
        },
        'MIGRATION': {
            'SKIP': False,  # alwys ignore migrations, changing this most likely breaks compatibility!
            'YES': False,  # always confirm migrations
            'BACKUP': False,  # always backup the database before running migrations
            'BACKUP_TMP': '{collection}_backup_{time}',  # format for backup collections
        },
        'CANCEL_TIMEOUT': 60,  # wait up to 60s for canceling an experiment
        'CONFIG_RESOLUTION_PROGRESS_BAR_THRESHOLD': 25,
        'AUTOCOMPLETE_CACHE_ALIVE_TIME': 60 * 60 * 24,  # one day
        'SETUP_COMMAND': '',
        'END_COMMAND': '',
        'SSH_FORWARD': {
            'LOCK_FILE': '/tmp/seml_ssh_forward.lock',
            'RETRIES_MAX': 6,
            'RETRIES_DELAY': 1,
            'LOCK_TIMEOUT': 30,
            'HEALTH_CHECK_INTERVAL': 10,
        },
    },
)

# Load user settings
if SETTINGS.USER_SETTINGS_PATH.exists():
    user_settings_source = run_path(str(SETTINGS.USER_SETTINGS_PATH))
    SETTINGS = munchify(merge_dicts(SETTINGS, user_settings_source['SETTINGS']))

SETTINGS.SLURM_STATES.ACTIVE = (
    SETTINGS.SLURM_STATES.PENDING
    + SETTINGS.SLURM_STATES.RUNNING
    + SETTINGS.SLURM_STATES.PAUSED
)
